import numpy as np
import cv2
from collections import deque

from util import *
from func import encrypt, batch_encrypt, divide


def solve_t_y(p, c, z):
    """calculate I* and Y"""
    t, y = np.zeros_like(c), np.zeros_like(c)
    for i in range(1, z.size):
        t[i] = c[z[i]] ^ mp(c[z[i - 1]], z[i])
    t[0] = c[z[0]] ^ mp(t[-1], z[0])
    for i in range(1, z.size):
        y[i] = t[i] ^ p[z[i]] ^ mp(t[i - 1], z[i])
    t = t.reshape(-1, 2).swapaxes(0, 1).reshape(-1)
    y = y.reshape(-1, 2).swapaxes(0, 1).reshape(-1)
    return t, y


def search_n0(p, y, z):
    """brute search N0"""
    p = p[z.reshape(-1, 2).swapaxes(0, 1).reshape(-1)]
    for n0 in range(1, y.size // 2):
        y[0], ns, nk = y[n0], 0, n0
        while (y[ns:ns + nk] == y[ns + nk:ns + 2 * nk]).all():
            ns, nk = ns + nk, int(p[ns:ns + nk].mean()) + nk
            if ns + 2 * nk > y.size:
                return n0
    raise Exception('unable to find N0')


def solve_N0_C(p, c, z):
    t, y = solve_t_y(p, c, z)
    n0 = search_n0(p, y, z)
    c_in = ms(t[0] ^ y[0] ^ p[z[0]], z[0])
    return n0, c_in


def get_longest_y(p255, c255, z, n0):
    """obtain Y^L"""
    _, y = solve_t_y(p255, c255, z)
    length = divide(p255, n0)
    sts = np.cumsum(length)
    ind = np.argmax(length)
    st, ed = sts[ind - 1], sts[ind]
    return y[st:ed]


def search(a, p0, p1, c0, c1):
    """z'(b) = a, find z'(b-1)"""
    eq = c0[a] ^ c1[a] == p0[a] ^ p1[a] ^ mp(c0, a) ^ mp(c1, a)
    can = [x for x in np.argwhere(eq.all(1)).flat if x != a]
    assert can, "b=0, attack failed"
    assert len(can) == 1, "unable to find z'(b-1), attack failed"
    return can[0]


def search_back(z, c0, c1, n):
    """find z(i) for i=(b+1)~(HW-1)"""
    cx = c0 ^ c1
    ind = np.arange(n, dtype=np.int32)[:, None]
    visit = np.ones(n, dtype=bool)
    visit[z[0]] = visit[z[1]] = False
    while True:
        x1, x2 = z[-2], z[-1]
        # get I^*
        I0, I1 = (c0[x2] ^ mp(c0[x1], x2)), (c1[x2] ^ mp(c1[x1], x2))
        eq = mp(I0, ind) ^ mp(I1, ind) ^ \
            mp(c0[x2], ind) ^ mp(c1[x2], ind) == cx
        can = [x for x in np.argwhere(eq.all(1)).flat if visit[x]]
        if len(can) == 0:
            return visit
        if len(can) > 1:
            print(len(z))
        assert len(can) == 1, "unable to recover z'"
        z.append(can[0])
        visit[z[-1]] = False


def search_front(z, c0, c1, visit):
    """find z(i) for i=0~(b-2)"""
    cx = c0 ^ c1
    visit[z[0]] = True
    cans, multi_cans = {}, {}
    for i in visit.nonzero()[0].flat:
        eq = cx[i] == mp(c0, i) ^ mp(c1, i)
        can = [x for x in np.argwhere(eq.all(1)).flat if x != i and visit[x]]
        if len(can) == 1:
            cans[i] = can[0]
        elif len(can) > 1:
            multi_cans[i] = can

    for value in cans.values():
        visit[value] = False
    flag = True
    while flag:
        for k, v in multi_cans.items():
            if len(v) == 1:
                continue
            v = multi_cans[k] = [x for x in v if visit[x]]
            if len(v) == 1:
                cans[k], visit[v[0]] = v[0], False
                break
        else:
            flag = False
    multi_cans = {k: v for k, v in multi_cans.items() if len(v) > 1}
    while z[0] in cans:
        z.appendleft(cans.pop(z[0]))
    return z


def generate_images(num, index):
    """generate chosen plain-image for z(b)=index"""
    generate_pair(index, num)
    batch_encrypt('data/gen/p', 'data/gen/c')
    encrypt('data/p255.bmp', dst='data')


def main():
    num = 5
    h = w = 256
    n = h * w
    st = np.random.randint(n)
    generate_images(num, st)

    p0, p1 = batch_read('data/gen/p/a', num), batch_read('data/gen/p/b', num)
    c0, c1 = batch_read('data/gen/c/a', num), batch_read('data/gen/c/b', num)
    p255, c255 = feed('data/p255.bmp'), feed('data/p255_e.bmp')

    z_real = random(n).argsort().astype(int)
    z_real_ = z_real.reshape(2, -1).swapaxes(0, 1).reshape(-1)
    z = deque([search(st, p0, p1, c0, c1), st])
    visit = search_back(z, c0, c1, n)
    z = np.array(search_front(z, c0, c1, visit))
    assert z.size == n and (z == z_real_).all(), f'only found {z.size}/{n}.'

    n0, c_in = solve_N0_C(p1[:, 0], c1[:, 0], z)
    y_longest = get_longest_y(p255, c255, z, n0)
    z = z.reshape(-1, 2).swapaxes(0, 1).reshape(-1)
    np.savetxt('data/z.txt', z, '%d')
    np.savetxt('data/y.txt', y_longest, '%d')
    print(f'result: n0={n0}, C={c_in}')


if __name__ == '__main__':
    main()
